/*
 * Copyright (c) 2005 Massachusetts Institute of Technology
 *                    All rights reserved
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */
/*
 * Functions for implementing SSH hopping as a single command line input.
 */

#include "includes.h"
RCSID("$OpenBSD$");

#include <ctype.h>
#include <errno.h>
#include <netdb.h>
#include <string.h>
#include <stdio.h>
#include <signal.h>
#include <stdlib.h>
#include <unistd.h>
#include <openssl/evp.h>
#include <openssl/err.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <sys/wait.h>
#include "log.h"
#include "misc.h"
#include "pathnames.h"
#include "ssh.h"
#include "readconf.h"
#include "xmalloc.h"
#include "sshhop.h"

/* Structure to encapsulate the information needed to represent each hop
 * in the chain connecting this host to the final destination. A hop is
 * an invocation of SSH that connects this host to an intermediary host or
 * a destination host, where there is at least one intermediary host between
 * this host (source) and the destination host.
 * command line syntax: ssh [ssh_options] -H [ssh_options] [user1@]host1 ...
 */
struct Hop {
	char	*user;		/* User name at remote host for this hop */
	char	*host;		/* Remote host to hop to (name or IP address) */
	int	address_type;	/* AF_INET | AF_INET6 (for "host" value) */
	pid_t	 pid;		/* Process id for spawned SSH subprocess */
	u_short  port;		/* SSH port number on this hop's remote host */
	int	 argc;		/* Number of command line options for this hop */
	char	**argv;		/* List of command line options for this hop */
	int	 cmdc;		/* Number of words in command to be invoked
				 * on this hop's remote host by ssh
				 */
	char	**cmdv;		/* Command to be invoked by SSH on this hop's
				 * remote host
				 */
};
typedef struct Hop Hop;
#define MAX_HOPS	10	/* Allow no more hops than this in the chain */

/* "Class" variables. We avoid passing these variables around as arguments
 * since 1) it simplifies the API in ssh.c and 2) it protects the integrity
 * of the information by limiting access to functions in this module.
 */
static Hop	s_hoplist[MAX_HOPS];	/* Information for each hop */
static int	s_numhops = 0;		/* Number of hops */
static char   **s_global_argv = NULL;	/* List of options that apply to all hops */
static int      s_global_argc = 0;	/* Number of options that apply to all hops */

/* External variables used for command line option parsing in getopt. */
extern int optind;
extern int optreset;

/* Local functions */
static int  sshop_get_option_list(int argc, char **argv, int starting_index,
    int *subset_argc, char ***subset_argv, int *hostloc, usage_fn *usage);
static int  sshop_parse_hop_info(char *input_str);
static void sshop_copy_global_args(Hop *hl, int hl_index, int g_index, int num);
static void sshop_process_global_options(usage_fn *usage);
static void sshop_process_local_options(Hop *hl, usage_fn *usage);
static int  sshop_set_hop_args(int i, char *host, struct sockaddr_un *tmp_uds,
    int max_argc, char **new_argv);
static int  sshop_start_ssh_client(char *argv[], Hop *hopinfo, char *uds_path);


/*========================= Functions used in ssh.c =========================*/

/* sshop_parse_options
 * Parse the command line options to process -H options
 * 
 * Inputs:
 *		argc	command line argument count (argc from input to main())
 *		argv	command line argument vector (argv from input to main())
 *		usage	function to print command usage information
 * Outputs:
 *		s_global_argv
 *		s_global_argc
 *		s_hoplist
 *		s_numhops
 * Returns:	number of hops
 *		-1 if error
 */
int
sshop_parse_options(int argc, char **argv, usage_fn *usage)
{
	int   new_argc, local_argc = 0;		/* argument counts */
	char **local_argv = NULL;		/* temp storage for arg list */
	int   hops = 0;				/* return variable */
	int   hop;				/* loop index */
	int   hoploc;				/* position in arg list of -H */

	/* Get the global options (options before the first -H option or EOL) */
	new_argc = sshop_get_option_list(argc, argv, 1, &s_global_argc,
				&s_global_argv, &hoploc, usage);

	/* While options remain, get information for next hop */
	while (new_argc > 0) {
		/* Make sure there is room for more hops */
		if (s_numhops == MAX_HOPS) {
			hops = -1;
			break;
		}
		/* Get option list for the next hop (to the next -H or EOL) */
		argv += (argc - new_argc);
		argc = new_argc;
		local_argv = NULL;
		local_argc = 0;
		new_argc = sshop_get_option_list(argc, argv, 0, &local_argc,
				&local_argv, &hoploc, usage);

		/* Process and store information about the next hop */
		if ((local_argc > 0) && (-1 < hoploc)) {
			/* Parse the user/hostname string */
			hop = sshop_parse_hop_info(argv[hoploc]);
			/* If user and host was successfully loaded (no parse
			 * errors), store the option information.
			 */
			if (hop > 0) {
				Hop *hl = &s_hoplist[s_numhops];
				hl->argc = hoploc;
				/* Commands follow the options for this hop */
				if (hl->argc > 0) {
					hl->argv = local_argv;
					hl->cmdc = local_argc - hoploc - 1;
				/* No commands follow options for this hop */
				} else if (local_argc > 0) {
					hl->argv = local_argv;
					hl->cmdc = 0;
				/* No options for this hop */
				} else if (local_argv) {
					xfree(local_argv);
					hl->argv = NULL;
					hl->cmdc = 0;
				}
				/* Number of command words > 0, store command */
				if (hl->cmdc > 0) {
					hl->cmdv = local_argv + hoploc + 1;
				/* Otherwise, no command */
				} else {
					hl->cmdv = NULL;
				}
				local_argv = NULL;
				s_numhops++;
			/* Otherwise, user/host parse error. End processing */
			} else {
				hops = -1;
				break;
			}
		/* The beginning of the option list for this host must be -H.
		 * If it isn't, move on to the next host but indicate an error.
		 */
		} else if (strcmp(argv[0], "-H") != 0) {
			hops = -1;
		}
	}

	/* Too many hops -- error */
	if (s_numhops == MAX_HOPS) {
		fprintf(stderr, "Limit of %d hops\n", MAX_HOPS-1);
		hops = -1;
	/* No errors and multiple hops. Prepare the hop option lists. */
	} else if (hops == 0 && 1 < s_numhops) {
		for (hop = 0; hop < s_numhops; hop++) {
			sshop_process_local_options(&s_hoplist[hop], usage);
		}
		sshop_process_global_options(usage);
		hops = s_numhops;
	/* No multiple hops, thus no need for any additional preparation of the
	 * option lists. Free function memory and return. */
	} else if (local_argv != NULL) {
		xfree(local_argv);
		local_argv = NULL;
	}

	return(hops);
}

/* sshop_set_default_user
 * Set the value of the user field for each hop where it is not already set.
 *
 * Inputs:
 *		user		default user name
 *		s_hoplist
 *		s_numhops
 * Outputs:
 *		s_hoplist (changes user value)
 * Returns:	none
 */
void
sshop_set_default_user(char *user)
{
	int i;
	Hop *hl;

	if (user == NULL)
		return;

	for (i = 0; i < s_numhops; i++) {
		hl = &s_hoplist[i];
		if (hl->user == NULL) {
			hl->user = xstrdup(user);
		}
	}
}

/* sshop_build_chain
 * Build the hop chain by starting each SSH subprocess in turn
 *
 * Inputs:
 *		ssh_path	path to SSH executable
 *		s_global_argc
 *		s_hoplist
 *		s_numhops
 * Outputs:	none
 * Returns:	 0 if all hops established successfully
 *		-1 otherwise
 */
int
sshop_build_chain(char *ssh_path)
{
	char  *localhost = "localhost";
	char  *host = s_hoplist[0].host;/* Source end of all connections */
	char  **new_argv;		/* Argument list used for ssh call */
	int   new_argc = 0;		/* Number of args for single ssh call */
	int   max_argc = 0;		/* Length of new_argv */
	int   i;			/* Loop counter */
	int   error = 0;		/* Return variable */
	struct sockaddr_un tmp_uds;	/* Unix domain socket */

	/* (Over-)Allocate space for the new, larger, argument list
	 * Number of elements is the sum of:
	 *	 #command line args specific to this hop
	 *	 #words in command
	 *	 max # options added by rote (hard-coded value)
	 *	 #command line args applied globally
	 */
	for (i = 0; i < s_numhops; i++) {
		new_argc = s_hoplist[i].argc + s_hoplist[i].cmdc;
		if (max_argc < new_argc) {
			max_argc = new_argc;
		}
	}
	max_argc += (15 + s_global_argc);

	/* Expand the size of the arg list and copy the arguments */
	new_argv = (char **)xmalloc(max_argc*sizeof(char *)); 
	/* Set the command arg: <ssh_path> */
	if (new_argv) {
		new_argv[0] = ssh_path;
	} else {
		exit(1);
	}

	memset(&tmp_uds, '\0', sizeof(tmp_uds));

	/* Create a connection for each hop in the chain */
	for (i = 0; (i < s_numhops) && (error == 0); i++) {
		/* Fill in the argument list for this hop */
		new_argc = sshop_set_hop_args(i, host, &tmp_uds,
					max_argc, new_argv);

		/* Start a new SSH session */
		if (new_argc > 5) {
			error = sshop_start_ssh_client(new_argv,
					&s_hoplist[i], tmp_uds.sun_path);
		} else {
			error = -1;
		}

		/* Reset for the next pass */
		host = localhost;
	}

	return(error);
}

/* sshop_kill_subprocesses
 * Kill the subprocesses. Doesn't guarantee that all processes have been
 * killed. (See sshop_wait_for_exit())
 *
 * Inputs:
 *		s_hoplist
 *		s_numhops
 * Outputs:
 *		s_hoplist (changes pid value)
 * Returns:	0  if all processes killed successfully
 *		1  if some processes may still be running
 */
int
sshop_kill_subprocesses()
{
	int   i;		/* Loop counter */
	pid_t pid;		/* Process id of killed process */
	int   looping = 0;	/* Loop status */
	int   status = 0;	/* Wait status */
	Hop   *hl;		/* Hop pointer */

	/* Debugging printout */
	debug("%d hop processes to kill", s_numhops);
	for (i = 0; i < s_numhops; i++) {
		hl = &s_hoplist[i];
		debug("hop %d: host: %s  pid: %d  port: %d",
		    i, hl->host, hl->pid, hl->port);
	}

	for (i = 0; i < s_numhops; i++) {
		hl = &s_hoplist[i];
		/* Send a TERM signal to running process */
		if (hl->pid > 0) {
			kill(hl->pid, SIGTERM);
			sleep(1);
			pid = waitpid(hl->pid, &status, WNOHANG);
			if (pid == hl->pid) {
				printf("process %d exited (host %s)\n",
				    pid, hl->host);
				hl->pid = 0;
			} else {
				looping = 1;
			}
		}
	}
	return(looping);
}

/* sshop_wait_for_exit
 * Wait for any one of the subprocesses to exit and then clean up the rest.
 * Doesn't return until all processes are killed and reaped.
 *
 * Inputs:
 *		s_hoplist
 *		s_numhops
 * Outputs:
 *		s_hoplist (changes pid value)
 * Returns:	none
 */
void
sshop_wait_for_exit()
{
	int   i;		/* Loop counter */
	pid_t pid;		/* Process id for process to kill */
	int   looping = 0;	/* Whether to keep looping or not */
	int   status = 0;	/* Wait status */
	Hop   *hl;		/* Pointer to hop info */

	/* Debugging printout */
	debug("%d hop process settings", s_numhops);
	for (i = 0; i < s_numhops; i++) {
		hl = &s_hoplist[i];
		debug("hop %d: host: %s  pid: %d  port: %d",
		    i, hl->host, hl->pid, hl->port);
	}

	/* Make sure there are processes to wait for */
	for (i = 0; (i < s_numhops) && !looping; i++) {
		looping = s_hoplist[i].pid != 0;
	}

	/* Wait for all subprocesses to exit */
	while (looping) {
		pid = wait(&status);
		for (i = 0; (i < s_numhops) && (pid != s_hoplist[i].pid); i++);

		/* If any process in the hop chain has exited, kill the rest */
		if (i < s_numhops) {
			printf("process %d exited (host %s)\n", pid,
			    s_hoplist[i].host);
			s_hoplist[i].pid = 0;
			looping = sshop_kill_subprocesses();
		}
	}
}

/* sshop_cleanup
 * Free memory associated with hop processing.
 *
 * Inputs:
 *		s_global_argv
 *		s_global_argc
 *		s_hoplist
 *		s_numhops
 * Outputs:
 *		s_global_argv
 *		s_global_argc
 *		s_hoplist
 *		s_numhops
 * Returns:	None
 */
void
sshop_cleanup()
{
	int i;
	Hop *hl;

	/* Free the list of arguments applied to all hops */
	if (s_global_argv) {
		xfree(s_global_argv);
		s_global_argv = NULL;
	}
	s_global_argc = 0;

	/* Free the information associated with each hop */
	for (i = 0; i < s_numhops; i++) {
		hl = &s_hoplist[i];
		if (hl->user) {
			xfree(hl->user);
			hl->user = NULL;
		}
		if (hl->host) {
			xfree(hl->host);
			hl->host = NULL;
		}
		if (hl->argv) {
			xfree(hl->argv);
			hl->argv = NULL;
		}
		hl->argc = 0;
		hl->cmdc = 0;
	}
	s_numhops = 0;
}

/*====================== Internal functions =============================*/

/*
 * Store the command line options for a single subsection. Each
 * subsection ends with either -H or '\0' (for the last subsection).
 *
 * Inputs:
 *		argc		command line argument count
 *		argv		command line argument vector
 *		starting_index	which argv to start with (usually 0 or 1)
 *		usage		function to print command usage information
 *		s_global_argc
 * Outputs:
 *		subset_argc	number of args in subsection
 *		subset_argv	list of args in subsection
 *		hostloc		location of host name in arg list
 * Returns:	number of arguments remaining to be processed
 */
static int
sshop_get_option_list(int argc, char **argv, int starting_index,
    int *subset_argc, char ***subset_argv, int *hostloc, usage_fn *usage)
{
	int  size;		/* Memory allocation size */
	int  hop_not_found = 1;	/* Loop terminator = -H or \0 not yet found */
	char ch;		/* Option value */
	int  new_argc = 0;	/* Return value */
	int  next_arg = 1;	/* Argument list index */
	int  i;			/* Loop counter */

	/* Initialize */
	*subset_argv = NULL;
	*subset_argc = 0;
	*hostloc = -1;
	optind   = starting_index;
	optreset = 1;

	/* Since the hop info includes hostname and possibly command
	 * information, 
	 */
	while (hop_not_found &&
		(ch = getopt(argc, argv, SSH_CMDLINE_OPTIONS)) != -1) {
		switch (ch) {
		/* Hop info, -H signals beginning of next hop information
		 * We use it here to signal the end of the previous set of
		 * ssh options. If there is a hostname, we will not find
		 * the -H before exiting the option list. Thus, this case
		 * will be used only for finding the end of the global options.
		 */
		case 'H':
			new_argc = argc - optind;
			*subset_argc = optind - 1;
			*hostloc = optind - 2;
			hop_not_found = 0;
			break;
		/* We'll copy all the other options later, keep track of
		 * where we are.
		 */
		default:
			if (ch != '?') {
				next_arg = optind;
			} else {
				usage();
			}
			break;
		}
	}
	/* -H not found yet (probably hit a hostname first), continue looking */
	if (hop_not_found && (optind <= argc)) {
		*hostloc = optind;
		for (i = next_arg; (i < argc) && (strcmp(argv[i], "-H") != 0);
		     i++);
		if (i <= argc) {
			new_argc = argc - i;
			*subset_argc  = i;
			hop_not_found = 0;
		}
	}
	/* Copy the arguments up to the next hop info or the end of the
	 * args, whichever came first. We will eventually use the allocated
	 * arg list for storing global arguments as well as hop specific
	 * arguments, so size it accordingly. Note that the first time
	 * this function is called, it will read in the global arguments.
	 * So the global arg list will not have extra space allocated.
	 */
	if ((s_global_argc + *subset_argc) > 0) {
		size = (s_global_argc + *subset_argc) * sizeof(*argv);
		*subset_argv = (char **)xmalloc(size);
		memset((void *)*subset_argv, '\0', size);
		if (*subset_argc > 0) {
			memcpy((void *)*subset_argv, (void *)argv,
				*subset_argc * sizeof(*argv));
		}
	}

	return(new_argc);
}

/* sshop_parse_hop_info
 * Parse the hop argument and store the username and hostname in
 * the hop structure. Uses s_numhops but does not change its value.
 * Format of input_str: [<user>@]<host>
 *
 * Inputs:
 *		input_str	string of the form [<user>@]<host>
 *		s_hoplist
 *		s_numhops
 * Outputs:
 *		s_hoplist
 * Returns:	number of hops, including this one, on success
 *		-1 on failure
 */
static int
sshop_parse_hop_info(char *input_str)
{
	char *p, buf[512];	/* if buf size changes, change the sscanf
				 * format statement
				 */
	Hop  *hl = NULL;	/* Ptr to hop information */
	char *hopstr;		/* Temporary string with user/host */
	int  hops = 0;		/* return variable */

	hopstr = xstrdup(input_str);
	/* Proceed only if the string is valid */
	if (hopstr && (*hopstr != '\0') && (s_numhops < MAX_HOPS)) {
		/* Skip over whitespace at beginning of string */
		while (isspace(*hopstr))
			hopstr++;
		/* If there are characters left, parse the string
		 * The string should have the format: [user@]host
		 */
		if (*hopstr != '\0') {
			memset(buf, '\0', sizeof(buf));
			/* Nothing scanned: Bad format */
			if (sscanf(hopstr, "%511c", buf) < 1) {
				fprintf(stderr,
				    "Bad SSH hop specification '%s'\n",
				    hopstr);
				hops = -1;
			/* Otherwise, input formatted ok, store the information.
			 * Don't increment hop count here. Just store the info.
			 */
			} else {
				hl = &s_hoplist[s_numhops];
				hl->pid = 0;
				/* User name specified */
				if ((p = strrchr(buf, '@'))) {
					*p = '\0';
					hl->user = xstrdup(buf);
					hl->host = xstrdup(++p);
				/* No user name specified */
				} else {
					hl->user = NULL;
					hl->host = xstrdup(buf);
				}
				hl->port = 0;
				/* Host string looks like IPv6 address */
				if (strchr(hl->host, ':')) {
					hl->address_type = AF_INET6;
				/* Otherwise, assume IPv4 address */
				} else {
					hl->address_type = AF_INET;
				}
				/* Set return value */
				hops = s_numhops + 1;
			}
		}
	}

	xfree(hopstr);

	return(hops);
}

/* sshop_copy_global_args
 * Copy consecutive arguments from the global list to the beginning of
 * the local list. Assumes the array was sized to allow addition of
 * the max number of global arguments.
 *
 * Inputs:
 *		hl		pointer to hop list item
 *		hl_index	location in arg list to copy to
 *		ga_index	index of the last global arg to be copied
 *		num		num of global arguments to copy
 *		s_global_argv
 *		s_global_argc
 * Outputs:
 *		hl (changes argv and argc values)
 * Returns:	none
 */
static void
sshop_copy_global_args(Hop *hl, int hl_index, int ga_index, int num)
{
	int i, j;

	if (hl->argv && (0 < num) && (num < s_global_argc)) {
		/* Make space for the global arguments at the indicated
		 * position in the hop argument list.
		 */
		bcopy(hl->argv+hl_index, hl->argv+num+hl_index,
		    (hl->argc-hl_index)*sizeof(*hl->argv));
		/* Copy in the global args. Note that the address is
		 * being copied, not the contents.
		 */
		for (i=hl_index, j=ga_index-num; i < hl_index+num; i++, j++) {
			hl->argv[i] = s_global_argv[j];
			hl->argc++;
		}
	} else {
		fatal("sshop_copy_global_args: no argument list to copy");
	}
}

/* sshop_process_global_options
 * Examine each command line option to be associated with all hops and
 * copy the options into the individual hop option lists, as appropriate.
 * Assumes that the individual option lists were sized large enough
 * to hold all of the global options, in addition to the local options.
 *
 * Inputs:
 *		usage		function to print command usage information
 *		s_global_argv
 *		s_global_argc
 *		s_hoplist
 *		s_numhops
 * Outputs:
 *		s_global_argv
 *		s_global_argc
 *		s_hoplist	(possibly update arguments)
 * Returns:	none
 */
static void
sshop_process_global_options(usage_fn *usage)
{
	extern Options options;		/* from ssh.c */
	extern int     debug_flag;	/* from ssh.c */

	int     i, j;		/* loop counters */
	u_short port;		/* port number specified as global option */
	char    ch;		/* option variable */
	Hop     *hl;		/* pointer to information for a single hop */
	int     hl_index0 = 0, hl_indexn = 0;
	optind   = 1;		/* option index for getopt */
	optreset = 1;		/* getopt to treat this as a new option list */

	/* Loop to process the global option list. The options will be
	 * added to the argument lists of the appropriate hops.
	 */
	while ((ch = getopt(s_global_argc, s_global_argv, SSH_CMDLINE_OPTIONS)) != -1) {
		switch (ch) {
		/* All hops, no arguments */
		case '1':
		case '2':
		case '4':
		case '6':
		case 'a':
		case 'C':
		case 'k':
		case 'P':
		case 'q':
			/* Retain these options */
			break;

		/* Verbosity, retain this option for subprocesses and apply
		 * it to the main process as well.
		 */
		case 'v':
			if (debug_flag == 0) {
				debug_flag = 1;
				options.log_level = SYSLOG_LEVEL_DEBUG1;
			} else {
				if (options.log_level < SYSLOG_LEVEL_DEBUG3)
					options.log_level++;
				break;
			}
			/* FALLTHROUGH */

		/* Version, retain this option for subprocesses and apply
		 * it to the main process as well.
		 */
		case 'V':
			fprintf(stderr, "%s, %s\n",
			    SSH_RELEASE, SSLeay_version(SSLEAY_VERSION));
			break;

		/* All hops, with arguments */
		case 'c':
		case 'F':
		case 'i':
		case 'I':
		case 'm':
		case 'o':
			/* Retain these options */
			break;

		/* Applies to all hops, copy the name */
		case 'l':
			for (i = 0; i < s_numhops; i++) {
				if (s_hoplist[i].user == 0) {
					s_hoplist[i].user = xstrdup(optarg);
				}
			}
			s_global_argv[optind-2] = NULL;
			s_global_argv[optind-1] = NULL;
			break;

		/* Use as default when port not specified */
		case 'p':
			port = a2port(optarg);
			if (0 < port) {
				for (i = 0; i < s_numhops; i++) {
					if (s_hoplist[i].port == 0) {
						s_hoplist[i].port = port;
					}
				}
			}
			s_global_argv[optind-2] = NULL;
			s_global_argv[optind-1] = NULL;
			break;

		/* First hop only, no arguments
		 * Prepend this option to the list of options */
		case 'b':
			hl = &s_hoplist[0];
			sshop_copy_global_args(hl, hl_index0, optind, 1);
			hl_index0++;
			s_global_argv[optind-1] = NULL;
			break;

		/* First hop only, with arguments 
		 * Prepend this option to the list of options */
		case 'u':
			hl = &s_hoplist[0];
			sshop_copy_global_args(hl, hl_index0, optind, 2);
			hl_index0 += 2;
			s_global_argv[optind-2] = NULL;
			s_global_argv[optind-1] = NULL;
			break;

		/* Last hop only, no arguments 
		 * Prepend this option to the list of options */
		case 'A':
		case 'f':
		case 'g':
		case 'n':
		case 'M':
		case 'N':
		case 's':
		case 't':
		case 'T':
		case 'x':
		case 'X':
		case 'Y':
			hl = &s_hoplist[s_numhops-1];
			sshop_copy_global_args(hl, hl_indexn, optind, 1);
			hl_indexn++;
			s_global_argv[optind-1] = NULL;
			break;

		/* Last hop only, with arguments 
		 * Prepend this option to the list of options */
		case 'D':
		case 'e':
		case 'L':
		case 'O':
		case 'R':
		case 'S':
		case 'U':
			hl = &s_hoplist[s_numhops-1];
			sshop_copy_global_args(hl, hl_indexn, optind, 2);
			hl_indexn += 2;
			s_global_argv[optind-2] = NULL;
			s_global_argv[optind-1] = NULL;
			break;

		default:
			usage();
			break;
		}
	}

	/* Remove all the nulls (from deleted args) from the global argv list
	 * and update the global arg count
	 */
	i = 0;
	while (i < s_global_argc) {
		for (; i < s_global_argc && s_global_argv[i] != NULL; i++);
		for (j = i; j < s_global_argc && s_global_argv[j] == NULL; j++);
		if (i < s_global_argc && j < s_global_argc) {
			bcopy(s_global_argv+j, s_global_argv+i,
				(s_global_argc - j) * sizeof(*s_global_argv));
			s_global_argc -= (j - i);
		} else if (i < s_global_argc && j >= s_global_argc) {
			s_global_argc = i;
		}
	}
}

/* sshop_process_local_options
 * Examine each command line option associated with a single hop and
 * determine whether the option should be retained or not
 *
 * Inputs:
 *		hl	hop for which to process options
 *		usage	function to print command usage information
 *		s_hoplist
 *		s_numhops
 * Outputs:
 *		hl	(updates argv and argc)
 * Returns:	none
 */
static void
sshop_process_local_options(Hop *hl, usage_fn *usage)
{
	int  i, j;	/* Loop counters */
	char ch;	/* option value */
	optind   = 0;	/* option index to start with */
	optreset = 1;	/* getopt to treat this as a new option list */

	/* Check for valid input */
	if (hl == NULL) {
		return;
	}
	if (hl->argc == 0) {
		return;
	}

	/* Loop to process the local option list for a single hop. Delete
	 * any options that are disallowed for this hop (based on whether
	 * it is the first, last, or an intermediate hop).
	 */
	while ((ch = getopt(hl->argc, hl->argv, SSH_CMDLINE_OPTIONS)) != -1) {
	switch (ch) {
		/* All hops, no arguments */
		case '1':
		case '2':
		case '4':
		case '6':
		case 'a':
		case 'A':
		case 'b':
		case 'C':
		case 'k':
		case 'P':
		case 'q':
		case 'v':
		case 'V':
			/* Retain these options */
			break;

		/* All hops, with arguments */
		case 'c':
		case 'e':
		case 'F':
		case 'i':
		case 'I':
		case 'm':
		case 'o':
			/* Retain these options */
			break;

		/* Copy the name into an empty user field. Whether copied
		 * or not, delete the option since it is extraneous.
		 */
		case 'l':
			if (hl->user == NULL) {
				hl->user = xstrdup(optarg);
			}
			hl->argv[optind-1] = NULL;
			hl->argv[optind-2] = NULL;
			break;

		/* First hop only, with arguments */
		case 'u':
			/* If not the first hop, delete this option and its
			 * argument. */
			if (hl != &s_hoplist[0]) {
				hl->argv[optind-1] = NULL;
				hl->argv[optind-2] = NULL;
			}
			break;

		/* Last hop only, no arguments */
		case 'f':
		case 'g':
		case 'n':
		case 'M':
		case 'N':
		case 'p':
		case 's':
		case 't':
		case 'T':
		case 'x':
		case 'X':
		case 'Y':
			/* If not the last hop, delete this option */
			if (hl != &s_hoplist[s_numhops-1]) {
				hl->argv[optind-1] = NULL;
			}
			break;

		/* Last hop only, with arguments */
		case 'D':
		case 'L':
		case 'O':
		case 'R':
		case 'S':
		case 'U':
			/* If not the last hop, delete this option and its
			 * argument. */
			if (hl != &s_hoplist[s_numhops-1]) {
				hl->argv[optind-1] = NULL;
				hl->argv[optind-2] = NULL;
			}
			break;

		default:
			usage();
			break;
	}
	}

	/* Remove all the nulls (from deleted options) from the local
	 * argv list and update the arg count
	 */
	i = 0;
	while (i < hl->argc) {
		for (; i < hl->argc && hl->argv[i] != NULL; i++);
		for (j = i; j < hl->argc && hl->argv[j] == NULL; j++);
		if (i < hl->argc && j < hl->argc) {
			bcopy(hl->argv+j, hl->argv+i,
				(hl->argc - j) * sizeof(*hl->argv));
			hl->argc -= (j - i);
		} else if (i < hl->argc && j >= hl->argc) {
			hl->argc = i - 1;
		}
	}

}

/* sshop_set_hop_args
 * Non-re-entrant function
 * Fill in the argument list for the next ssh command
 * Returns the number of arguments in the argument list -- should be > 5
 * for a valid option list.
 *
 * Inputs:
 *		i 		hop index for this hop
 *		host
 *		tmp_uds 	Unix domain socket used for the last hop
 *		max_argc	maximum length of the argument list
 *		new_argv	argument list that will be used to start ssh
 *		s_global_argv
 *		s_global_argc
 *		s_hoplist	(contains ssh path as first arg in arg list)
 *		s_numhops
 * Outputs:
 *		tmp_uds 	Unix domain socket used for this hop
 *		s_hoplist	(argv and argc updated)
 * Returns:	Number of arguments in the argument list
 */
static int
sshop_set_hop_args(int i, char *host, struct sockaddr_un *tmp_uds,
    int max_argc, char **new_argv)
{
	static char  *c_bckgnd_flag    = "-n";
		/* Don't use -f since we don't want to fork a new process */
	static char  *c_no_cmd_flag    = "-N";
	static char  *c_force_tty_flag = "-t";
	static char  *c_no_forward_x11 = "-x";
	static char  tmp_port_arg[8];
	static char  tmp_fwd_arg[372];
	static char  tmp_host_alias[272];
	static struct sockaddr_un tmp_uds_last;

	char  tmp_dst_host_arg[1024];
	int   new_argc;			/* return variable */
	pid_t pid = getpid();
	char  *user = NULL;
	char  delimiter = ',';
	struct servent *sp;		/* Used to get the SSH port # */
	int   j = i + 1;

	/*----- Set up -----*/
	/* Save the last value to be used later */
	memcpy(tmp_uds_last.sun_path, tmp_uds->sun_path,
	    sizeof(tmp_uds_last.sun_path));

	/* Generate the argument to the -U option for forwarding.
	 * Since -U is not used for the last ssh invocation (no
	 * more forwarding), we only  generate the argument for
	 * s_numhops - 1 hops.
	 */
	if (j < s_numhops) {
		/* If needed, set the remote ssh port number */
		if (s_hoplist[j].port == 0) {
			sp = getservbyname(SSH_SERVICE_NAME, "tcp");
			if (sp)
				s_hoplist[j].port = ntohs(sp->s_port);
			else
				s_hoplist[j].port = SSH_DEFAULT_PORT;
		}
		/* Get the user name on the remote host */
		if (s_hoplist[j].user) {
			user = xstrdup(s_hoplist[j].user);
		} else {
			user = NULL;
		}
		/* Create a name for the unix domain socket that
		 * will be used for forwarding.
		 */
		construct_uds_path(NULL, s_hoplist[j].host, pid,
			&user, s_hoplist[j].port,
			tmp_uds->sun_path, sizeof(tmp_uds->sun_path));
		/* Create the argument in the form:
		 * local_path,remote_host,remote_port
		 */
		snprintf(tmp_fwd_arg, sizeof(tmp_fwd_arg), "%s%c%s%c%d",
			tmp_uds->sun_path, delimiter, s_hoplist[j].host,
			delimiter, s_hoplist[j].port);
		/* Clean up */
		if (user)
			xfree(user);

	/* Last hop in the chain, zero out values to signal that
	 * they shouldn't be used. */
	} else {
		memset(tmp_fwd_arg, '\0', sizeof(tmp_fwd_arg));
		memset(tmp_uds->sun_path, '\0',
			sizeof(tmp_uds->sun_path));
	}

	/* Construct the host argument (will follow all other options
	 * and precede command)
	 */
	if (s_hoplist[i].user) {
		snprintf(tmp_dst_host_arg, sizeof(tmp_dst_host_arg),
			"%s@%s", s_hoplist[i].user, host);
	} else {
		strncpy(tmp_dst_host_arg, host,
			sizeof(tmp_dst_host_arg));
	}
	/* -o HostKeyAlias= argument */
	if (strcmp(host, "localhost") == 0) {
		snprintf(tmp_host_alias, sizeof(tmp_host_alias),
			"HostKeyAlias=%s", s_hoplist[i].host);
	} else {
		memset(tmp_host_alias, '\0', sizeof(tmp_host_alias));
	}

	/*--- Start building the argument list ----*/

	/* Copy in the new arguments. Do something different for the last
	 * hop in the chain since it will be an interactive session. */
	new_argc = 1;
	if (j < s_numhops) {
		/* ssh -n -N -X */
		new_argv[new_argc++] = c_bckgnd_flag;
		new_argv[new_argc++] = c_no_cmd_flag;
		new_argv[new_argc++] = c_no_forward_x11;
	} else {
		/* ssh -t -t */
		new_argv[new_argc++] = c_force_tty_flag;
		new_argv[new_argc++] = c_force_tty_flag;
	}
	/* Tell ssh to use the key for the real hostname
	 * -o HostKeyAlias=<hostname> */
	if (strlen(tmp_host_alias) > 0) {
		new_argv[new_argc++] = "-o";
		new_argv[new_argc++] = tmp_host_alias;
	}
	/* Port info for the first hop only
	 * [-p <SSH_port_number>] */
	if (i == 0 && 0 < s_hoplist[i].port) {
		snprintf(tmp_port_arg, sizeof(tmp_port_arg), "%d",
			s_hoplist[i].port);
		new_argv[new_argc++] = "-p";
		new_argv[new_argc++] = tmp_port_arg;
	}
	/* SSH forwarding information
	 * -U <uds_path>,<remote_host>,<remote_SSH_port> */
	if (strlen(tmp_fwd_arg) > 0) {
		new_argv[new_argc++] = "-U";
		new_argv[new_argc++] = tmp_fwd_arg;
	}
	/* Unix domain socket to connect to for SSH connection
	 * -u <last_hop_uds_path> */
	if (strlen(tmp_uds_last.sun_path) > 0) {
		new_argv[new_argc++] = "-u";
		new_argv[new_argc++] = tmp_uds_last.sun_path;
	}
	/* Copy in the global options, if any. Skip the first
	 * since that is the program path
	 * ssh <fixed_options> [<global_ssh_options>]
	 */
	if (s_global_argc > 1) {
		memcpy(new_argv+new_argc, s_global_argv+1,
			(s_global_argc-1) * sizeof(char *));
		new_argc += (s_global_argc - 1);
	}
	/* Copy in the options specific to this host, if any
	 * ssh <fixed_options> [<global_ssh_options>]
	 *     [<hop_specific_ssh_options>]
	 */
	if (s_hoplist[i].argc > 0) {
		memcpy(new_argv+new_argc, s_hoplist[i].argv,
	    		s_hoplist[i].argc * sizeof(char *));
		new_argc += s_hoplist[i].argc;
	}
	/* Copy in the destination host
	 * ssh <fixed_options> [<global_ssh_options>]
	 *     [<hop_specific_ssh_options>] [<user>@]<host>
	 */
	new_argv[new_argc++] = tmp_dst_host_arg;
	/* Copy in any commands
	 * ssh <fixed_options> [<global_ssh_options>]
	 *     [<hop_specific_ssh_options>] [<user>@]<host> [<commmands>]
	 */
	if (s_hoplist[i].cmdc > 0) {
		memcpy(new_argv+new_argc, s_hoplist[i].cmdv,
			s_hoplist[i].cmdc * sizeof(char *));
		new_argc += s_hoplist[i].cmdc;
	}
	/* Terminate argument list */
	new_argv[new_argc++] = NULL;

	return(new_argc);
}

/* sshop_start_ssh_client
 * Spawn a subprocess
 * Inputs:
 *		argv - null terminated argument vector for invoking ssh
 * 		hopinfo - information about this hop
 *		uds_path - path to the unix domain socket. The existence
 *			   of this socket is monitored to determine when
 *			   ssh process has made the remote connection.
 * Outputs:
 * 		hopinfo (pid updated)
 * Returns:	 0 if the remote connection appears to have been established
 *		-1 otherwise
 */
static int
sshop_start_ssh_client(char *argv[], Hop *hopinfo, char *uds_path)
{
	pid_t pid;		/* return value from fork() */
	int   error = 0;	/* return variable */
	int   status;		/* wait status */

	pid = fork();
	/* Child process; exec the command */
	if (pid == 0) {
		/* Should never return from this call */
		execv(argv[0], argv);
		/* If we do return, kill off the process immediately. */
		fatal("sshop_start_ssh_client: exec(%s): %s", argv[0],
		    strerror(errno));
	/* Parent; store the pid for the child */
	} else if (pid > 0) {
		/* Store the process id for the child so we can kill it later */
		hopinfo->pid = pid;

		/* Make sure the process didn't have issues at startup */
		if (waitpid(pid, &status, WNOHANG) == pid) {
			hopinfo->pid = 0;
			return -1;
		}

		/* Put up notification that SSH subprocess is starting */
		if (hopinfo->user != NULL) {
			fprintf(stdout, "[%d]\n[%d] login for %s@%s\n",
			    hopinfo->pid, hopinfo->pid, hopinfo->user,
			    hopinfo->host);
			fflush(stdout);
		} else {
			fprintf(stdout, "[%d]\n[%d] login to %s\n",
			    hopinfo->pid, hopinfo->pid, hopinfo->host);
			fflush(stdout);
		}
		/* Wait for the process to establish the connection
		 * If there is a problem, we will sit here indefinitely.
		 * The user will need to kill the process (Ctl-C)
		 * Since the last process in the chain does not create
		 * a unix domain socket, it will return immediately.
		 */
		if (uds_path && (strlen(uds_path) > 0)) {
			while (access(uds_path, F_OK) == -1 && error == 0) {
				/* Make sure the process is still running */
				pid = hopinfo->pid;
				if (waitpid(pid, &status, WNOHANG) == pid) {
					hopinfo->pid = 0;
					error = -1;
				} else {
					sleep(1);
				}
			}
			sleep(1);
		}
	/* Error */
	} else {
		fprintf(stderr,
		    "sshop_start_ssh_client: Unable to fork child process\n");
		error = -1;
	}

	return(error);
}
